import re

from .base import RegexBasedDetector


class SquareOAuthDetector(RegexBasedDetector):
    """Scans for Square OAuth Secrets"""
    secret_type = 'Square OAuth Secret'

    denylist = [
        re.compile(r'sq0csp-[0-9A-Za-z\\\-_]{43}'),
    ]



"""
This plugin searches for Mailchimp keys
"""
import re
from base64 import b64encode

import requests

from ..constants import VerifiedResult
from .base import RegexBasedDetector


class MailchimpDetector(RegexBasedDetector):
    """Scans for Mailchimp keys."""
    secret_type = 'Mailchimp Access Key'

    denylist = (
        re.compile(r'[0-9a-z]{32}-us[0-9]{1,2}'),
    )

    def verify(self, secret: str) -> VerifiedResult:  # pragma: no cover
        _, datacenter_number = secret.split('-us')

        response = requests.get(
            'https://us{}.api.mailchimp.com/3.0/'.format(
                datacenter_number,
            ),
            headers={
                'Authorization': b'Basic ' + b64encode(
                    'any_user:{}'.format(secret).encode('utf-8'),
                ),
            },
        )
        return (
            VerifiedResult.VERIFIED_TRUE
            if response.status_code == 200
            else VerifiedResult.VERIFIED_FALSE
        )



from enum import Enum
from os import getenv
from sys import stdout


def supports_ansi_colors() -> bool:
    return (getenv('CLICOLOR', '1') != '0' and stdout.isatty())\
        or getenv('CLICOLOR_FORCE', '0') != '0'


class AnsiColor(Enum):
    RESET = '[0m'
    BOLD = '[1m'
    RED = '[91m'
    RED_BACKGROUND = '[41m'
    LIGHT_GREEN = '[92m'
    PURPLE = '[95m'


def colorize(text: str, color: AnsiColor) -> str:
    if not supports_ansi_colors():
        return text

    return '\x1b{}{}\x1b{}'.format(
        color.value,
        text,
        AnsiColor.RESET.value,
    )



import hashlib
import inspect


def get_caller_path(offset: int = 0) -> str:
    """
    This is a utility function to get the caller's fully qualified python import path,
    so that it can be used to index into the global settings object. It is highly
    recommended to cache the response to this (for performance reasons), as such:

    >>> @lru_cache(maxsize=1)
    ... def _get_specific_caller_path() -> str:
    ...     return get_caller_path(offset=1)

    For a deeper dive into why this performance matters, check out
    https://stackoverflow.com/a/17366561/13340678, and estimate how many secrets you will
    need to filter out (and thereby, invoke this function for)

    :raises: IndexError
    """
    stack = inspect.stack()
    frame_info = stack[1 + offset]      # +1 because we don't want the current frame.

    module_path = frame_info.frame.f_globals['__name__']
    function_name = frame_info.function
    return f'{module_path}.{function_name}'


def compute_file_hash(filename: str, buffer_size: int = 64 * 1024) -> str:
    """
    When we make any modifications to the inputs of a baseline, the baseline should
    also reflect these changes. Otherwise, we can get into a strange situation of
    irreproducibility: a "hidden" change to the underlying model may produce different
    results.

    To ensure this doesn't happen, we capture a hash of the model that we are using.
    This way, if the model changes, the baseline will have to change, and we will be
    able to better track the changes this way.

    This is akin to:
        $ sha1sum <filename>
    """
    sha1 = hashlib.sha1()  # noqa: S324
    with open(filename, 'rb') as f:
        data = f.read(buffer_size)
        while data:
            sha1.update(data)
            data = f.read(buffer_size)

    return sha1.hexdigest()



from __future__ import annotations

from typing import Any
from typing import Set
from typing import Union

import requests

from ..constants import VerifiedResult
from ..core.potential_secret import PotentialSecret
from ..util.code_snippet import CodeSnippet
from .base import RegexBasedDetector
from .high_entropy_strings import Base64HighEntropyString


class IbmCloudIamDetector(RegexBasedDetector):
    """Scans for IBM Cloud IAM Key."""

    secret_type = 'IBM Cloud IAM Key'
    # opt means optional
    opt_ibm_cloud_iam = r'(?:ibm(?:_|-|)cloud(?:_|-|)iam|cloud(?:_|-|)iam|' + \
        r'ibm(?:_|-|)cloud|ibm(?:_|-|)iam|ibm|iam|cloud|)'
    opt_dash_underscore = r'(?:_|-|)'
    opt_api = r'(?:api|platform|)'
    key_or_pass = r'(?:key|pwd|password|pass|token)'
    secret = r'([a-zA-Z0-9_\-]{44}(?![a-zA-Z0-9_\-]))'
    denylist = [
        RegexBasedDetector.build_assignment_regex(
            prefix_regex=opt_ibm_cloud_iam + opt_dash_underscore + opt_api,
            secret_keyword_regex=key_or_pass,
            secret_regex=secret,
        ),
    ]

    def __init__(self) -> None:
        super().__init__()
        self.entropy_plugin = Base64HighEntropyString()

    def verify(self, secret: str) -> VerifiedResult:
        response = verify_cloud_iam_api_key(secret)

        return VerifiedResult.VERIFIED_TRUE if response.status_code == 200 \
            else VerifiedResult.VERIFIED_FALSE

    def analyze_line(
        self,
        filename: str,
        line: str,
        line_number: int = 0,
        context: CodeSnippet | None = None,
        raw_context: CodeSnippet | None = None,
        **kwargs: Any,
    ) -> Set[PotentialSecret]:
        potentials = super().analyze_line(
            filename, line, line_number, context, raw_context, **kwargs,
        )
        secrets = set()
        for p in potentials:
            if self.entropy_plugin.analyze_line(
                    filename, f'"{p.secret_value}"', line_number, context, raw_context, **kwargs,
            ):
                secrets.add(p)
        return secrets


def verify_cloud_iam_api_key(apikey: Union[str, bytes]) -> requests.Response:  # pragma: no cover
    if isinstance(apikey, bytes):
        apikey = apikey.decode('UTF-8')

    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Accept': 'application/json',
    }
    response = requests.post(
        'https://iam.cloud.ibm.com/identity/token',
        headers=headers,
        data={
            'grant_type': 'urn:ibm:params:oauth:grant-type:apikey',
            'apikey': apikey,
        },
    )
    return response



import contextlib
from contextlib import contextmanager
from copy import deepcopy
from functools import lru_cache
from importlib import import_module
from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from urllib.parse import urlparse

from .exceptions import InvalidFile
from .util.importlib import import_file_as_module


@lru_cache(maxsize=1)
def get_settings() -> 'Settings':
    """
    This is essentially a singleton pattern, that allows for (controlled) global access
    to common variables.
    """
    return Settings()


def configure_settings_from_baseline(baseline: Dict[str, Any], filename: str = '') -> 'Settings':
    """
    :raises: KeyError
    """
    settings = get_settings()

    if 'plugins_used' in baseline:
        settings.configure_plugins(baseline['plugins_used'])

    if 'filters_used' in baseline:
        settings.configure_filters(baseline['filters_used'])

        if 'detect_secrets.filters.wordlist.should_exclude_secret' in settings.filters:
            config = settings.filters['detect_secrets.filters.wordlist.should_exclude_secret']

            from detect_secrets import filters
            filters.wordlist.initialize(
                wordlist_filename=config['file_name'],
                min_length=config['min_length'],
                file_hash=config['file_hash'],
            )

        if 'detect_secrets.filters.gibberish.should_exclude_secret' in settings.filters:
            config = settings.filters['detect_secrets.filters.gibberish.should_exclude_secret']

            from detect_secrets import filters
            filters.gibberish.initialize(
                model_path=config.get('model'),
                limit=config['limit'],
            )

    if filename:
        settings.filters['detect_secrets.filters.common.is_baseline_file'] = {
            'filename': filename,
        }

    return settings


@contextmanager
def default_settings() -> Generator['Settings', None, None]:
    """Convenience function to enable all plugins and default filters."""
    from .core.plugins.util import get_mapping_from_secret_type_to_class

    with transient_settings({
        'plugins_used': [
            {'name': plugin_type.__name__}
            for plugin_type in get_mapping_from_secret_type_to_class().values()
        ],
    }) as settings:
        yield settings


@contextmanager
def transient_settings(config: Dict[str, Any]) -> Generator['Settings', None, None]:
    """Allows the customizability of non-global settings per invocation."""
    original_settings = get_settings().json()

    cache_bust()
    try:
        yield configure_settings_from_baseline(config)
    finally:
        cache_bust()
        configure_settings_from_baseline(original_settings)


def cache_bust() -> None:
    get_plugins.cache_clear()

    get_filters.cache_clear()
    for path in get_settings().filters:
        # Need to also clear the individual caches (e.g. cached regex patterns).
        parts = urlparse(path)
        if not parts.scheme:
            module_path, _ = path.rsplit('.', 1)
            try:
                module = import_module(module_path)
            except ModuleNotFoundError:
                continue
        elif parts.scheme == 'file':
            file_path = path[len('file://'):].split('::')[0]
            try:
                module = import_file_as_module(file_path)
            except (FileNotFoundError, InvalidFile):
                continue
        else:
            continue

        for item_key in dir(module):
            item = getattr(module, item_key)
            try:
                if item.__module__ != module_path:
                    # Make sure we only clear the cache specific to the module.
                    raise AttributeError

                item.cache_clear()
            except AttributeError:
                pass

    get_settings.cache_clear()


class Settings:
    DEFAULT_FILTERS = {
        'detect_secrets.filters.common.is_invalid_file',
        'detect_secrets.filters.heuristic.is_non_text_file',
    }

    def __init__(self) -> None:
        self.clear()

    def clear(self) -> None:
        # mapping of class names to initialization variables
        self.plugins: Dict[str, Dict[str, Any]] = {}

        # mapping of python import paths to configuration variables
        self.filters: Dict[str, Dict[str, Any]] = {
            path: {}
            for path in {
                *self.DEFAULT_FILTERS,
                'detect_secrets.filters.allowlist.is_line_allowlisted',
                'detect_secrets.filters.heuristic.is_sequential_string',
                'detect_secrets.filters.heuristic.is_potential_uuid',
                'detect_secrets.filters.heuristic.is_likely_id_string',
                'detect_secrets.filters.heuristic.is_templated_secret',
                'detect_secrets.filters.heuristic.is_prefixed_with_dollar_sign',
                'detect_secrets.filters.heuristic.is_indirect_reference',
                'detect_secrets.filters.heuristic.is_lock_file',
                'detect_secrets.filters.heuristic.is_not_alphanumeric_string',
                'detect_secrets.filters.heuristic.is_swagger_file',
            }
        }

    def set(self, other: 'Settings') -> None:
        self.plugins = other.plugins
        self.filters = other.filters

    def configure_plugins(self, config: List[Dict[str, Any]]) -> 'Settings':
        """
        :param config: e.g.
            [
                {'name': 'AWSKeyDetector'},
                {'limit': 4.5, 'name': 'Base64HighEntropyString'}
            ]
        """
        for plugin in config:
            plugin = {**plugin}
            name = plugin.pop('name')
            self.plugins[name] = plugin

        get_plugins.cache_clear()
        return self

    def disable_plugins(self, *plugin_names: str) -> 'Settings':
        for name in plugin_names:
            with contextlib.suppress(KeyError):
                self.plugins.pop(name)

        get_plugins.cache_clear()
        return self

    def configure_filters(self, config: List[Dict[str, Any]]) -> 'Settings':
        """
        :param config: e.g.
            [
                {'path': 'detect_secrets.filters.heuristic.is_sequential_string'},
                {
                    'path': 'detect_secrets.filters.regex.should_exclude_files',
                    'pattern': '^test.*',
                }
            ]
        """
        self.filters = {
            path: {}
            for path in self.DEFAULT_FILTERS
        }

        # Make a copy, so we don't affect the original.
        filter_configs = deepcopy(config)
        for filter_config in filter_configs:
            path = filter_config['path']
            self.filters[path] = filter_config

        get_filters.cache_clear()
        return self

    def disable_filters(self, *filter_paths: str) -> 'Settings':
        for filter_path in filter_paths:
            self.filters.pop(filter_path, None)

        get_filters.cache_clear()
        return self

    def json(self) -> Dict[str, Any]:
        plugins_used = []
        for plugin in get_plugins():
            # NOTE: We use the initialized plugin's JSON representation (rather than using
            # the configured settings) to deal with cases where plugins define their own
            # default variables, that is not necessarily carried through the
            # settings object.
            serialized_plugin = plugin.json()

            plugins_used.append({
                # We want this to appear first.
                'name': serialized_plugin['name'],

                # NOTE: We still need to use the saved settings configuration though, since
                # there are keys specifically in the settings object that we need to carry over
                # (e.g. `path` for custom plugins).
                **self.plugins[serialized_plugin['name']],

                # Finally, this comes last so that it overrides any values that are saved in
                # the settings object.
                **serialized_plugin,
            })

        return {
            'plugins_used': sorted(
                plugins_used,
                key=lambda x: str(x['name'].lower()),
            ),
            'filters_used': sorted(
                [
                    {
                        'path': path,
                        **config,
                    }
                    for path, config in self.filters.items()
                    if path not in self.DEFAULT_FILTERS
                ],
                key=lambda x: str(x['path'].lower()),
            ),
        }


@lru_cache(maxsize=1)
def get_plugins() -> List:
    # We need to import this here, otherwise it will result in a circular dependency.
    from .core import plugins

    return [
        plugins.initialize.from_plugin_classname(classname)
        for classname in get_settings().plugins
    ]


@lru_cache(maxsize=1)
def get_filters() -> List:
    from .core.log import log
    from .util.inject import get_injectable_variables

    output = []
    for path in get_settings().filters:
        parts = urlparse(path)
        if not parts.scheme:
            module_path, function_name = path.rsplit('.', 1)
            try:
                function = getattr(import_module(module_path), function_name)
            except (ModuleNotFoundError, AttributeError):
                log.warning(f'Invalid filter: {path}')
                continue

        elif parts.scheme == 'file':
            file_path, function_name = path[len('file://'):].split('::')

            try:
                function = getattr(import_file_as_module(file_path), function_name)
            except (FileNotFoundError, InvalidFile, AttributeError):
                log.warning(f'Invalid filter: {path}')
                continue

        else:
            log.warning(f'Invalid filter: {path}')
            continue

        # We attach this metadata to the function itself, so that we don't need to
        # compute it everytime. This will allow for dependency injection for filters.
        function.injectable_variables = set(get_injectable_variables(function))
        output.append(function)

        # This is for better logging.
        function.path = path

    return output



import argparse
import inspect
import os
from importlib import import_module
from urllib.parse import urlparse

from ... import filters
from ...constants import VerifiedResult
from ...core.log import log
from ...exceptions import InvalidFile
from ...settings import get_settings
from ...util.importlib import import_file_as_module
from .common import valid_path


def add_filter_options(parent: argparse.ArgumentParser) -> None:
    parser = parent.add_argument_group(
        title='filter options',
        description=(
            'Configure settings for filtering out secrets after they are flagged '
            'by the engine.'
        ),
    )

    verify_group = parser.add_mutually_exclusive_group()
    verify_group.add_argument(
        '-n',
        '--no-verify',
        action='store_true',
        help='Disables additional verification of secrets via network call.',
    )
    verify_group.add_argument(
        '--only-verified',
        action='store_true',
        help='Only flags secrets that can be verified.',
    )

    parser.add_argument(
        '--exclude-lines',
        type=str,
        action='append',
        help='If lines match this regex, it will be ignored.',
    )
    parser.add_argument(
        '--exclude-files',
        type=str,
        action='append',
        help='If filenames match this regex, it will be ignored.',
    )
    parser.add_argument(
        '--exclude-secrets',
        type=str,
        action='append',
        help='If secrets match this regex, it will be ignored.',
    )

    if filters.wordlist.is_feature_enabled():
        parser.add_argument(
            '--word-list',
            type=valid_path,
            help=(
                'Text file with a list of words, '
                'if a secret contains a word in the list we ignore it.'
            ),
            dest='word_list_file',
        )

    if filters.gibberish.is_feature_enabled():
        parser.add_argument(
            '--gibberish-model',
            type=valid_path,
            help='Path to model trained with gibberish-detector.',
            dest='gibberish_model_file',
        )
        parser.add_argument(
            '--gibberish-limit',
            type=float,
            help='Threshold to determine whether a string is gibberish.',
        )

    _add_custom_filters(parser)
    _add_disable_flag(parser)


def _add_custom_filters(parser: argparse._ArgumentGroup) -> None:
    def valid_looking_paths(path: str) -> str:
        # Expected path format:
        #   - detect_secrets.filters.common.is_invalid_file (python import path)
        #   - testing/custom_filters.py::is_invalid_secret (local file)
        #   - file://testing/custom_filters.py::is_invalid_secret (local file)
        parts = urlparse(path)
        if not parts.scheme and '::' in path:
            # This could be a local file, without the file schema.
            path = 'file://' + path
            parts = urlparse(path)

        if parts.scheme == 'file':
            # May be local file.
            # We do some initial pre-processing, but perform the file validation during the
            # post-processing step.
            components = parts.path.split('::')
            if len(components) != 2:
                raise argparse.ArgumentTypeError(
                    'Did not specify function name for imported file.',
                )

            file_path = path[len('file://'):].split('::')[0]
            if not os.path.isfile(file_path):
                raise argparse.ArgumentTypeError(f'{file_path} is not a valid file.')
        elif parts.scheme:
            raise argparse.ArgumentTypeError(f'{path} is not a valid filter path.')

        return path

    parser.add_argument(
        '-f',
        '--filter',
        type=valid_looking_paths,
        nargs=1,
        action='append',        # so we can support multiple flags with same value
        help=(
            'Specify path to custom filter. '
            'May be a python module path (e.g. detect_secrets.filters.common.is_invalid_file) or '
            'a local file path (e.g. file://path/to/file.py::function_name).'
        ),
    )


def _add_disable_flag(parser: argparse._ArgumentGroup) -> None:
    parser.add_argument(
        '--disable-filter',
        type=str,
        nargs=1,
        action='append',        # so we can support multiple flags with same value
        help='Specify filter to disable. e.g. detect_secrets.filters.common.is_invalid_file',
    )


def parse_args(args: argparse.Namespace) -> None:
    if args.exclude_lines:
        get_settings().filters['detect_secrets.filters.regex.should_exclude_line'] = {
            'pattern': args.exclude_lines,
        }

    if args.exclude_files:
        get_settings().filters['detect_secrets.filters.regex.should_exclude_file'] = {
            'pattern': args.exclude_files,
        }

    if args.exclude_secrets:
        get_settings().filters['detect_secrets.filters.regex.should_exclude_secret'] = {
            'pattern': args.exclude_secrets,
        }

    if (
        filters.wordlist.is_feature_enabled()
        and args.word_list_file
    ):
        filters.wordlist.initialize(args.word_list_file)

    if filters.gibberish.is_feature_enabled():
        kwargs = {}
        if args.gibberish_model_file:
            kwargs['model_path'] = args.gibberish_model_file

        if args.gibberish_limit:
            kwargs['limit'] = args.gibberish_limit

        filters.gibberish.initialize(**kwargs)

    if not args.no_verify:
        get_settings().filters[
            'detect_secrets.filters.common.is_ignored_due_to_verification_policies'
        ] = {
            'min_level': (
                VerifiedResult.VERIFIED_TRUE
                if args.only_verified
                else VerifiedResult.UNVERIFIED
            ).value,
        }
    else:
        get_settings().disable_filters(
            'detect_secrets.filters.common.is_ignored_due_to_verification_policies',
        )

    if args.disable_filter:
        # Flatten entry for easier parsing.
        args.disable_filter = [entry for item in args.disable_filter for entry in item]

        redundant_disabled_filters = set(args.disable_filter) - set(get_settings().filters)
        for name in redundant_disabled_filters:
            log.warning(f'Redundant --disable-filter "{name}"')

        get_settings().disable_filters(*args.disable_filter)

    if args.filter:
        # Flatten entry for easier parsing.
        args.filter = [entry for item in args.filter for entry in item]

        # Post-processing validation
        for item in args.filter:
            _raise_if_custom_filter_path_is_invalid(item)
            get_settings().filters[item] = {}


def _raise_if_custom_filter_path_is_invalid(path: str) -> None:
    """Performs post-validation for custom filters."""
    parts = urlparse(path)
    if not parts.scheme:
        try:
            module_path, function_name = path.rsplit('.', 1)
        except ValueError:
            raise argparse.ArgumentTypeError(
                'Invalid Python module path for custom filter.',
            )

        try:
            module = import_module(module_path)
        except ModuleNotFoundError:
            raise argparse.ArgumentTypeError(f'Cannot import "{path}" as custom filter.')

        try:
            function = getattr(module, function_name)
        except AttributeError:
            raise argparse.ArgumentTypeError(
                f'No filter function named `{function_name}` found in "{module_path}".',
            )

        if not inspect.isfunction(function):
            raise argparse.ArgumentTypeError(f'{path} is not a filter function.')

    elif parts.scheme == 'file':
        file_path, function_name = path[len('file://'):].split('::')

        try:
            module = import_file_as_module(file_path)
        except (FileNotFoundError, InvalidFile):
            raise argparse.ArgumentTypeError(
                f'Cannot import {file_path} as custom filter.',
            )

        try:
            getattr(module, function_name)
        except AttributeError:
            raise argparse.ArgumentTypeError(
                f'No filter function named `{function_name}` found in "{file_path}".',
            )



import hashlib
import hmac
from datetime import datetime
from datetime import timezone
from typing import List

import requests

from ..constants import VerifiedResult
from ..util.code_snippet import CodeSnippet
from .base import RegexBasedDetector


class IbmCosHmacDetector(RegexBasedDetector):
    """Scans for IBM Cloud Object Storage HMAC credentials."""
    # requires 3 factors
    #
    #   access_key: access_key_id
    #   secret_key: secret_access_key
    #   host, defaults to 's3.us.cloud-object-storage.appdomain.cloud'

    secret_type = 'IBM COS HMAC Credentials'

    token_prefix = r'(?:(?:ibm)?[-_]?cos[-_]?(?:hmac)?|)'
    password_keyword = r'(?:secret[-_]?(?:access)?[-_]?key)'
    password = r'([a-f0-9]{48}(?![a-f0-9]))'
    denylist = (
        RegexBasedDetector.build_assignment_regex(
            prefix_regex=token_prefix,
            secret_keyword_regex=password_keyword,
            secret_regex=password,
        ),
    )

    def verify(  # type: ignore[override]
        self,
        secret: str,
        context: CodeSnippet,
    ) -> VerifiedResult:
        key_id_matches = find_access_key_id(context)

        if not key_id_matches:
            return VerifiedResult.UNVERIFIED

        try:
            for key_id in key_id_matches:
                verify_result = verify_ibm_cos_hmac_credentials(key_id, secret)
                if verify_result:
                    return VerifiedResult.VERIFIED_TRUE
        except requests.exceptions.RequestException:
            return VerifiedResult.UNVERIFIED

        return VerifiedResult.VERIFIED_FALSE


def find_access_key_id(context: CodeSnippet) -> List[str]:
    key_id_keyword_regex = r'(?:access[-_]?(?:key)?[-_]?(?:id)?|key[-_]?id)'
    key_id_regex = r'([a-f0-9]{32})'

    regex = RegexBasedDetector.build_assignment_regex(
        prefix_regex=IbmCosHmacDetector.token_prefix,
        secret_keyword_regex=key_id_keyword_regex,
        secret_regex=key_id_regex,
    )

    return [
        match
        for line in context
        for match in regex.findall(line)
    ]


def hash(key: bytes, msg: str) -> bytes:
    return hmac.new(key, msg.encode('utf-8'), hashlib.sha256).digest()


def createSignatureKey(key: str, datestamp: str, region: str, service: str) -> bytes:  # noqa: N802
    key_date = hash(('AWS4' + key).encode('utf-8'), datestamp)
    key_string = hash(key_date, region)
    key_service = hash(key_string, service)
    key_signing = hash(key_service, 'aws4_request')
    return key_signing


def verify_ibm_cos_hmac_credentials(
    access_key: str,
    secret_key: str,
    host: str = 's3.us.cloud-object-storage.appdomain.cloud',
) -> bool:
    response = query_ibm_cos_hmac(access_key, secret_key, host)
    return response.status_code == 200


def query_ibm_cos_hmac(
    access_key: str,
    secret_key: str,
    host: str = 's3.us.cloud-object-storage.appdomain.cloud',
) -> requests.Response:
    # Sample code referenced from link below
    # https://cloud.ibm.com/docs/services/cloud-object-storage/api-reference?topic=cloud-object-storage-hmac-signature  # noqa: E501

    # request elements
    http_method = 'GET'
    # region is a wildcard value that takes the place of the AWS region value
    # as COS doesn't use the same conventions for regions, this parameter can accept any string
    region = 'us-standard'
    endpoint = 'https://{}'.format(host)
    bucket = ''  # add a '/' before the bucket name to list buckets
    object_key = ''
    request_parameters = ''

    # assemble the standardized request
    time = datetime.now(timezone.utc)
    timestamp = time.strftime('%Y%m%dT%H%M%SZ')
    datestamp = time.strftime('%Y%m%d')

    standardized_resource = '/' + bucket + '/' + object_key
    standardized_querystring = request_parameters
    standardized_headers = 'host:' + host + '\n' + 'x-amz-date:' + timestamp + '\n'
    signed_headers = 'host;x-amz-date'
    payload_hash = hashlib.sha256(b'').hexdigest()

    standardized_request = (
        http_method + '\n'
        + standardized_resource + '\n'
        + standardized_querystring + '\n'
        + standardized_headers + '\n'
        + signed_headers + '\n'
        + payload_hash
    ).encode('utf-8')

    # assemble string-to-sign
    hashing_algorithm = 'AWS4-HMAC-SHA256'
    credential_scope = datestamp + '/' + region + '/' + 's3' + '/' + 'aws4_request'
    sts = (
        hashing_algorithm + '\n'
        + timestamp + '\n'
        + credential_scope + '\n'
        + hashlib.sha256(standardized_request).hexdigest()
    )

    # generate the signature
    signature_key = createSignatureKey(secret_key, datestamp, region, 's3')
    signature = hmac.new(
        signature_key,
        (sts).encode('utf-8'),
        hashlib.sha256,
    ).hexdigest()

    # assemble all elements into the 'authorization' header
    v4auth_header = (
        hashing_algorithm + ' '
        + 'Credential=' + access_key + '/' + credential_scope + ', '
        + 'SignedHeaders=' + signed_headers + ', '
        + 'Signature=' + signature
    )

    # create and send the request
    headers = {'x-amz-date': timestamp, 'Authorization': v4auth_header}
    # the 'requests' package automatically adds the required 'host' header
    request_url = endpoint + standardized_resource + standardized_querystring

    request = requests.get(request_url, headers=headers)

    return request



import argparse
from typing import cast

from . import baseline
from ...settings import get_settings
from .common import initialize_plugin_settings


def add_scan_action(parent: argparse._SubParsersAction) -> argparse.ArgumentParser:
    parser = parent.add_parser(
        'scan',
        help='Creates a baseline by scanning a repository for secrets.',
        description=(
            'Scans a repository for secrets in code. The generated output is compatible with '
            '`detect-secrets-hook --baseline`.'
        ),
    )

    _add_adhoc_scanning(parser)
    _add_pragma_scanning(parser)
    _add_initialize_baseline_options(parser)

    return cast(argparse.ArgumentParser, parser)


def _add_adhoc_scanning(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        '--string',
        nargs='?',
        const=True,
        help=(
            'Scans an individual string, and displays configured plugins\' verdict.'
        ),
    )


def _add_pragma_scanning(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        '--only-allowlisted',
        action='store_true',
        help=(
            'Only scans the lines that are flagged with `allowlist secret`. This helps '
            'verify that individual exceptions are indeed non-secrets.'
        ),
    )


def _add_initialize_baseline_options(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        'path',
        nargs='*',
        default=['.'],
        help=(
            'Scans the entire codebase and outputs a snapshot of '
            'currently identified secrets.'
        ),
    )

    group = parser.add_argument_group(title='scan options')
    group.add_argument(
        '--all-files',
        action='store_true',
        help='Scan all files recursively (as compared to only scanning git tracked files).',
    )
    baseline.add_baseline_option(
        cast(argparse.ArgumentParser, group),
        help='If provided, will update existing baseline by importing settings from it.',
    )
    group.add_argument(
        '--force-use-all-plugins',
        action='store_true',
        help=(
            'If a baseline is provided, detect-secrets will default to loading the plugins '
            'specified by that baseline. However, this may also mean it doesn\'t perform the '
            'scan with the latest plugins. If this flag is provided, it will always use the '
            'latest plugins'
        ),
    )
    group.add_argument(
        '--slim',
        action='store_true',
        help=(
            'Slim baselines are created with the intention of minimizing differences between '
            'commits. However, they are not compatible with the `audit` functionality, and '
            'slim baselines will need to be remade to be audited.'
        ),
    )


def parse_args(args: argparse.Namespace) -> None:
    if args.action != 'scan':
        return

    # NOTE: This is assumed to run *after* the baseline argument processor, and before
    # the plugin argument processor.
    if args.baseline is not None and args.force_use_all_plugins:
        get_settings().plugins.clear()
        initialize_plugin_settings(args)



import logging
import sys
from functools import partial
from typing import cast
from typing import Optional


def get_logger(name: Optional[str] = None, format_string: Optional[str] = None) -> 'CustomLogger':
    """
    :param name: used for declaring log channels.
    :param format_string: for custom formatting
    """
    logging.captureWarnings(True)
    log = logging.getLogger(name)

    # Bind custom method to instance.
    # Source: https://stackoverflow.com/a/2982
    log.set_debug_level = partial(CustomLogger.set_debug_level, log)    # type: ignore
    cast(CustomLogger, log).set_debug_level(0)

    # Setting up log formats
    log.handlers = []
    handler = logging.StreamHandler(sys.stderr)
    handler.setFormatter(
        logging.Formatter(
            format_string
            or
            '[%(module)s]\t%(levelname)s\t%(message)s',
        ),
    )
    log.addHandler(handler)

    return cast(CustomLogger, log)


def _set_debug_level(self: logging.Logger, debug_level: int) -> None:
    """
    :param debug_level: between 0-2, configure verbosity of log
    """
    mapping = {
        0: logging.ERROR,
        1: logging.INFO,
        2: logging.DEBUG,
    }

    self.setLevel(
        mapping[min(debug_level, 2)],
    )


class CustomLogger(logging.Logger):
    def set_debug_level(self, debug_level: int) -> None:
        """
        :param debug_level: between 0-2, configure verbosity of log
        """
        mapping = {
            0: logging.ERROR,
            1: logging.INFO,
            2: logging.DEBUG,
        }

        self.setLevel(
            mapping[min(debug_level, 2)],
        )


log = get_logger('detect-secrets')



"""
This plugin searches for Discord Bot Token
"""
import re

from .base import RegexBasedDetector


class DiscordBotTokenDetector(RegexBasedDetector):
    """Scans for Discord Bot token."""
    secret_type = 'Discord Bot Token'

    denylist = [
        # Discord Bot Token ([M|N|O]XXXXXXXXXXXXXXXXXXXXXXX[XX].XXXXXX.XXXXXXXXXXXXXXXXXXXXXXXXXXX)
        # Reference: https://discord.com/developers/docs/reference#authentication
        # Also see: https://github.com/Yelp/detect-secrets/issues/627
        re.compile(r'[MNO][a-zA-Z\d_-]{23,25}\.[a-zA-Z\d_-]{6}\.[a-zA-Z\d_-]{27}'),
    ]



class UnableToReadBaselineError(ValueError):
    """Think of this as a 404, if getting a baseline had a HTTPError code."""
    pass


class InvalidBaselineError(ValueError):
    """Think of this as a 400, if getting a baseline had a HTTPError code."""
    pass


class InvalidFile(ValueError):  # noqa: N818
    """Think of this as a 400, if FileNotFoundError was a 404 HTTPError code."""
    pass


class SecretNotFoundOnSpecifiedLineError(Exception):
    def __init__(self, line: int) -> None:
        super().__init__(
            'ERROR: Secret not found on line {}!\n'.format(line)
            + 'Try recreating your baseline to fix this issue.',
        )


class NoLineNumberError(Exception):
    def __init__(self) -> None:
        super().__init__(
            'ERROR: No line numbers found in baseline! Line numbers are needed '
            'for auditing secret occurrences. Try recreating your baseline to fix '
            'this issue.',
        )
